import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
import warnings
warnings.filterwarnings("ignore", message=".*Covariance of the parameters could.*")

def quadratic_polynomial_formula(x, a0, a1, a2):
    return np.exp(a0 + a1 * np.log(x) + a2 * np.log(x)**2)

def polyfit2(df, wavelength=550, x=None):
    # 提取要拟合的数据列
    y = df.to_numpy()
    params, _ = curve_fit(quadratic_polynomial_formula, x, y, check_finite=False, nan_policy='omit')
    # print(params)
    yfit = quadratic_polynomial_formula(wavelength, *params)
    return yfit

def group_arr_by_nan(arr):
    # 分离非nan的索引，返回字典
    index_dict = {}
    for idx, t in enumerate(arr):
        a = ",".join(map(str, np.where(~np.isnan(t))[0]))
        if a in index_dict:
            index_dict[a].append(idx)
        else:
            index_dict[a] = [idx]
    return index_dict

class LEVReader:
    def __init__(self, filepath, header=6) -> None:
        self.df = pd.read_csv(filepath, header=header, encoding="Windows 1252")
        self.df = self.df.replace(-999.0, np.nan)
        self.df = self._convert_time_to_utc(self.df)

    def filter_high_quality_aod_rows(self, dataframe=None):
        """筛选出质量较高的AOD数据行

        Parameters
        ----------
        dataframe : DataFrame
            数据源

        Returns
        -------
        DataFrame
            筛选后的数据
        """
        if dataframe is None:
            df = self.df.copy()
        else:
            df = dataframe.copy()
        df = df.query("AOD_440nm > 0.0")
        df = df.query("AOD_500nm > 0.0")
        df = df.query("AOD_675nm > 0.0")
        df = df.query("AOD_870nm > 0.0")
        return df
    
    def _convert_time_to_utc(self, dataframe):
        df = dataframe
        df.loc[:, "CustomDate"] = df.loc[:, "Date(dd:mm:yyyy)"] + " " + df.loc[:, "Time(hh:mm:ss)"]
        df.loc[:, "CustomDate"] = pd.to_datetime(df.loc[:, "CustomDate"], format='%d:%m:%Y %H:%M:%S', utc=True)
        df.set_index("CustomDate", inplace=True)
        df.sort_index(inplace=True)
        return df

    def _only_keep_aod_column(self, dataframe):
        df = dataframe
        # 删除非AOD以及无效的数据列
        columns = [title for title in df.columns if title.startswith('AOD_') and "Empty" not in title]
        df = df.loc[:, columns]
        df = df.replace(-999.0, np.nan)
        df = df.dropna(axis=1, how="all")
        # 将数据列名改为波长名
        columns = df.columns
        columns = [i.replace("AOD_", "").replace("nm", "") for i in columns]
        df.columns = columns
        # 根据波长大小对列名进行排序
        try:
            columns = [(int(i), i) for i in columns]
        except Exception as e:
            print("Aeronet数据格式与代码不同，请修改!")
            raise e
        columns = sorted(columns, key=lambda x: x[0])
        _, columns = zip(*columns)
        df = df.loc[:, columns]
        return df

    def interp_aod_xxxnm(self, dataframe, method="cubic", wavelength:float=550):
        if method in ['linear', 'nearest', 'nearest-up', 'zero_spline', 'slinear_spline', 'quadratic_spline', 'cubic_spline']:
            # 采用scipy的插值方法
            df_aod = self._only_keep_aod_column(dataframe.copy())
            x = np.array(list(map(int, df_aod.columns)))
            y = df_aod.to_numpy()
            method_real = method.replace("_spline", "") # 为了兼容scipy的插值参数格式, 去掉_spline后缀
            group_dict = group_arr_by_nan(y)
            series_list = []
            for key in group_dict:
                columns = list(map(int, key.split(",")))
                x_temp = x[columns]
                if (method_real == "cubic" and len(x_temp) < 4) or min(x_temp) > wavelength or max(x_temp) < wavelength:
                    yfit_temp = np.zeros(len(group_dict[key]))
                    yfit_temp[:] = np.nan
                    series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
                    continue
                y_temp = y[group_dict[key], :][:, columns]
                f = interp1d(x=x_temp, y=y_temp, kind=method_real, bounds_error=False)
                yfit_temp = f(wavelength)
                series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
            yfit = pd.concat(series_list).sort_index()
        elif method in ["scipy_curvefit"]:
            # https://blog.csdn.net/weixin_42182090/article/details/131018277
            from swifter import set_defaults

            set_defaults(progress_bar=False)
            df_aod = self._only_keep_aod_column(dataframe.copy())
            x = np.array(list(map(int, df_aod.columns)))
            yfit = df_aod.swifter.apply(polyfit2, axis=1, wavelength=wavelength, x=x)
        elif method in ["numpy_polyfit"]:
            df_aod = self._only_keep_aod_column(dataframe.copy())
            x = np.log(np.array(list(map(int, df_aod.columns))))
            y = np.log(df_aod.to_numpy().T)
            coef = np.polyfit(x, y, 2)
            yfit = np.exp(np.polyval(coef, np.log(wavelength)))
        elif method in ["Angstrom"]:
            df_raw = dataframe.copy()
            df_raw['440-675_Angstrom_Exponent'] = df_raw['440-675_Angstrom_Exponent'].astype(float)
            yfit = df_raw.apply(lambda x: x['AOD_440nm']*np.power((wavelength/440), -x['440-675_Angstrom_Exponent']), axis=1)
        elif method in ["cubic+scipy"]:
            yfit_scipy = self.interp_aod_xxxnm(dataframe=dataframe, method="scipy_curvefit", wavelength=wavelength)
            yfit_cubic = self.interp_aod_xxxnm(dataframe=dataframe, method="cubic_spline", wavelength=wavelength)
            yfit = (yfit_scipy + yfit_cubic)/2
        else:
            raise ValueError("method参数错误")
        return np.array(yfit)
    
    def compare(self, dataframe=None, filter_high_quality=True, wavelength=500):
        import time
        if dataframe is None:
            df_src = self.df.copy()
        else:
            df_src = dataframe.copy()
        df_src = df_src.query(f"AOD_{wavelength}nm > 0.0")
        if filter_high_quality:
            df_src = df_src.query("AOD_440nm > 0.0")
            df_src = df_src.query("AOD_500nm > 0.0")
            df_src = df_src.query("AOD_675nm > 0.0")
            df_src = df_src.query("AOD_870nm > 0.0")
        print("数据筛选与否: filter_high_quality = " + str(filter_high_quality))
        df = pd.DataFrame(index=df_src.index)
        df["true"] = df_src[f"AOD_{wavelength}nm"]
        df_src = df_src.drop(columns=[f"AOD_{wavelength}nm"])
        for method in ["Angstrom", "quadratic_spline", "cubic_spline", "scipy_curvefit", "cubic+scipy"]:
            name = method.split("_")[0]
            t_start = time.time()
            df[name] = self.interp_aod_xxxnm(dataframe=df_src, method=method, wavelength=wavelength)
            t_spended = time.time() - t_start
            bias = np.abs(df[name]-df["true"])
            print("{:15s}: 均误差{:10.6f}, 最大误差{:10.6f}, 耗时{:6.3f}s".format(name, bias.mean(), bias.max(), t_spended))
            
if __name__ == '__main__':
    path = r"E:\data-dbl\AOD_retrieval\AERONET\AERONET_lev15\19930101_20210102_Beijing_RADI.lev15"
    reader = LEVReader(path)
    reader.compare(filter_high_quality=True)


from mtmaod.aeronet import LEVReader


if __name__ == '__main__':
    # 读取数据
    lev_file_path = r"E:\data-dbl\AOD_retrieval\AERONET\AERONET_lev15\19930101_20210102_Beijing_RADI.lev15"
    reader = LEVReader(lev_file_path)

    # 筛选高质量数据
    df = reader.filter_high_quality_aod_rows()

    # 插值550nm波段的AOD, 当使用"scipy_curvefit" 或 "cubic+scipy"方法时, 调用多核CPU, 需要在"if __name__ == '__main__'"环境中使用
    df["AOD_550nm"] = reader.interp_aod_xxxnm(df, method="cubic_spline", wavelength=550)

    # 对比不同AOD插值方法在500nm波段的表现
    reader.compare(df, filter_high_quality=True, wavelength=500)

    # 更新reader实例的df变量
    reader.df = df

    # 只保留AOD列
    df_aod = reader._only_keep_aod_column(df)




import matplotlib as mpl
from .indicator import envelope_of_expected_error, root_mean_square_error, r2_score
import numpy as np
from mtmtool.io import auto_make_dirs
from matplotlib import pyplot as plt

style_dict = {
    "single_normal": {"figsize": (3.538, 3.538), "fontsize": 10.5, "ticksize": 9},
    "single_bold": {"figsize": (5, 4.5), "fontsize": 12, "ticksize": 10.5},
    # "double_normal": {"figsize": (6.7, 5)},
    # "double_bold": {"figsize": (7.48, 6)},
}

def density_chart(x, y, xlabel="AOD Measured", ylabel="AOD Retrieved", title=None, type="kernels", bins=25, xy_max_edge=2, save_path=None, dpi=300, figsize=(3, 3), style=None, fontsize=9, ticksize=8):
    """绘制密度图, 来比较标签和预测值的分布情况, 使用高斯核密度(kernels)来绘制密度图 或 使用直方图(scatter or grid)来统计单位面积内的点的数量

    Parameters
    ----------
    x : list
        label
    y : list
        predict
    xlabel : str, optional
        x轴标签, by default "AOD Measured"
    ylabel : str, optional
        y轴标签, by default "AOD Retrieved"
    title : str, optional
        标题, by default None
    type : str, {"kernels", "scatter", "grid"}, optional
        要绘制的图的类型, by default "kernels"
    bins : int, optional
        直方图统计时, 每1个单位有多少个箱子, 每个箱子单位刻度为(1/bins), by default 25
    xy_max_edge : int, optional
        最大的xy显示边界, by default 2
    save_path : _type_, optional
        保存路径, 如果为None, 则直接显示, by default None
    dpi : int, optional
        分辨率, 每英寸下多少个像元, by default 300
    figsize : tuple, optional
        画布大小, 单位是英寸, by default (3, 3)
    style : _type_, optional
        风格样式, by default None
    fontsize : int, optional
        标题等的字体大小, by default 9
    ticksize : int, optional
        tick等的字体大小, by default 8

    """
    if style in style_dict:
        figsize = style_dict[style].get("figsize", figsize)
        fontsize = style_dict[style].get("fontsize", fontsize)
        ticksize = style_dict[style].get("ticksize", ticksize)
    with mpl.rc_context():
        # 设置字体
        font = {'family': 'serif', 'serif': 'Times New Roman', 'weight': 'normal', 'size': fontsize}
        mpl.rc('font', **font)
        mpl.rcParams['mathtext.fontset'] = 'stix'

        # 设置布局
        mpl.rcParams["savefig.bbox"] = "tight"
        mpl.rcParams['figure.constrained_layout.use'] = True

        # Set x axis
        mpl.rcParams['xtick.direction'] = "in"
        mpl.rcParams['xtick.major.size'] = 3
        mpl.rcParams['xtick.major.width'] = 0.5
        mpl.rcParams['xtick.minor.size'] = 1.5
        mpl.rcParams['xtick.minor.width'] = 0.5
        mpl.rcParams['xtick.minor.visible'] = True
        mpl.rcParams['xtick.top'] = True

        # Set y axis
        mpl.rcParams['ytick.direction'] = "in"
        mpl.rcParams['ytick.major.size'] = 3
        mpl.rcParams['ytick.major.width'] = 0.5
        mpl.rcParams['ytick.minor.size'] = 1.5
        mpl.rcParams['ytick.minor.width'] = 0.5
        mpl.rcParams['ytick.minor.visible'] = True
        mpl.rcParams['ytick.right'] = True

        # 设置分辨率
        mpl.rcParams['figure.dpi'] = dpi  # plt.show显示分辨率
        mpl.rcParams["savefig.dpi"] = dpi # plt.savefig保存分辨率

        # 设置画布
        fig, axes = plt.subplots(1, 1, figsize=figsize)

        # 设置其它参数
        axes.set_xlabel(xlabel, labelpad=2)
        axes.set_ylabel(ylabel, labelpad=2)
        axes.set_title(title)
        axes.set_xlim(-0.05, xy_max_edge + 0.05)
        axes.set_ylim(-0.05, xy_max_edge + 0.05)
        axes.set_aspect("equal")
        axes.grid(linestyle=":", color="r", alpha=0.1)
        xyticks = [i/2 for i in range(int(xy_max_edge//0.5) + 1)]
        xyticks = [i for i in xyticks if i <= xy_max_edge]
        xytickslables = [f"{format(i, '.1f')}" for i in xyticks]
        axes.set_xticks(xyticks)
        axes.set_yticks(xyticks)
        axes.set_xticklabels(xytickslables)
        axes.set_yticklabels(xytickslables)
        axes.tick_params(labelsize=ticksize)

        # 将数据转为numpy数组
        x, y = np.array(x).reshape(-1), np.array(y).reshape(-1) # x means label, y means predict
        mask = (x <= xy_max_edge) & (y <= xy_max_edge)
        print(f"数据总数: {len(x)}, 有效数据总数: {mask.sum()}, 越界数据总数: {(~mask).sum()}")

        if type in ["kernels"]:
            from scipy.stats import gaussian_kde
            xy = np.vstack([x, y])
            z = gaussian_kde(xy)(xy)
            ax_data = axes.scatter(x, y, marker = 'o', c = z, s = 20, cmap = 'jet', vmin = 0)

        elif type in ["scatter", "grid"]:
            # 计算数据的直方图分布
            xy_max_edge = xy_max_edge
            nbins = xy_max_edge * bins
            x = x[mask]
            y = y[mask]
            H, xedges, yedges = np.histogram2d(x, y, bins=nbins, range=[[0, xy_max_edge], [0, xy_max_edge]])

            # 生成散点图图层
            if type == "scatter":
                x_index = (x * bins).astype("int")
                y_index = (y * bins).astype("int")
                z = H[x_index, y_index]
                # 获取散点大小
                axis_height = axes.xaxis.clipbox.height # 获取坐标轴高度
                axis_length = axes.get_xlim()[1] - axes.get_xlim()[0] # 获取坐标轴长度
                scatter_point_size = int(axis_height/(bins*(axis_length)))
                s = (scatter_point_size**2)/4
                # 画散点图
                sort_index = np.argsort(z)
                x = x[sort_index]
                y = y[sort_index]
                z = z[sort_index]
                ax_data = axes.scatter(x, y, marker='.', c=z, s=s, label="AOD scatter", cmap="jet", linewidths=0)
            elif type == "grid": # 绘制方格图
                # H needs to be rotated and flipped
                H = np.rot90(H)
                H = np.flipud(H)
                # Mask zeros
                Hmasked = np.ma.masked_where(H == 0, H)  # Mask pixels with a value of zero
                ax_data = axes.pcolormesh(xedges, yedges, Hmasked, cmap="jet", vmin=1, linewidth=0)
        else:
            raise ValueError(f"没有type: {type}参数对应的画图模式")
        
        # 生成包络线图层
        axes.plot([-1 / 3, xy_max_edge+1], [-1 / 3, (xy_max_edge+1) - (0.05 + 0.15 * (xy_max_edge+1))], linestyle='--', color='black', label='EE envelopes', linewidth = 0.5)
        axes.plot([-1 / 3, xy_max_edge], [-1 / 3, xy_max_edge + (0.05 + 0.15 * xy_max_edge)], linestyle='--', color='black', linewidth = 0.6)
        axes.plot([-1 / 3, xy_max_edge+1], [-1 / 3, xy_max_edge+1], linestyle='-', color='black', linewidth = 0.5)

        # 生成文本图层
        # # 计算指标
        above_percent, below_percent, within_percent = envelope_of_expected_error(x, y)
        rmse = root_mean_square_error(x, y)
        r2 = r2_score(x, y) if len(x) > 1 else 0
        add_text = "\n".join([
            f"$EE$ envelopes: ±(0.05+15%)",
            f"     {format(within_percent, '.1f')}% within $EE$",
            f"     {format(above_percent, '.1f')}% above $EE$",
            f"     {format(below_percent, '.1f')}% below $EE$",
            f"$RMSE$ = {format(rmse, '.3f')}",
            f"$R^2$ = {format(r2, '.3f')}",
            f"$N$ = {len(x)}",
        ])
        axes.text(x=0, y=xy_max_edge, s=add_text, fontsize=ticksize, verticalalignment='top', bbox=dict(boxstyle='round', facecolor='white', alpha=0, linewidth=0.2))

        # 生成图例
        lg = axes.legend(loc=4, alignment="right", fontsize=ticksize, labelspacing=0.2, borderpad=0.2, handletextpad=0.2, handlelength=1)
        lg.get_frame().set(linewidth=0.2, edgecolor="k", alpha=0.5)

        # 生成colorbar
        cb = fig.colorbar(ax_data, ax=axes, pad=0.02, fraction=0.05)
        if type in ["scatter", "grid"]:
            cb.ax.set_title(label='Num', loc='center', fontdict={"size": ticksize}, pad=0) #loc参数
            yticks = list(filter(lambda x: x == int(x), cb.ax.get_yticks()))
            cb.ax.set_yticks(yticks)
        cb.ax.tick_params(labelsize=ticksize)
        
        # 画图
        if save_path is not None:
            auto_make_dirs(save_path, isdir = False)
            fig.savefig(save_path, dpi=dpi)
            plt.close(fig)
        else:
            plt.show()